/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Public License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Public License, please send an email to 
 * dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Public License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System; using Microsoft;
using System.Collections.Generic;
using System.IO;
using Microsoft.Linq.Expressions;
using System.Reflection;
using Microsoft.Scripting.Actions;
using System.Text;

using Microsoft.Scripting.Utils;
using Microsoft.Scripting.Interpretation;

namespace Microsoft.Scripting.Runtime {
    /// <summary>
    /// Provides language specific facilities which are typicalled called by the runtime.
    /// </summary>
    public abstract class LanguageContext {
        private readonly ScriptDomainManager _domainManager;
        private static readonly ModuleGlobalCache _noCache = new ModuleGlobalCache(ModuleGlobalCache.NotCaching);
        private ActionBinder _binder;
        private readonly ContextId _id;

        protected LanguageContext(ScriptDomainManager domainManager) {
            ContractUtils.RequiresNotNull(domainManager, "domainManager");

            _domainManager = domainManager;
            _id = domainManager.GenerateContextId();
        }

        public ActionBinder Binder {
            get {
                return _binder;
            }
            protected set {
                _binder = value;
            }
        }

        /// <summary>
        /// Provides the ContextId which includes members that should only be shown for this LanguageContext.
        /// 
        /// ContextId's are used for filtering by Scope's.
        /// </summary>
        public ContextId ContextId {
            get { return _id; }
        }

        /// <summary>
        /// Gets the ScriptDomainManager that this LanguageContext is running within.
        /// </summary>
        public ScriptDomainManager DomainManager {
            get { return _domainManager; }
        }

        /// <summary>
        /// Whether the language can parse code and create source units.
        /// </summary>
        public virtual bool CanCreateSourceCode {
            get { return true; }
        }

        #region Scope

        public virtual Scope GetScope(string path) {
            return null;
        }

        public ScopeExtension GetScopeExtension(Scope scope) {
            ContractUtils.RequiresNotNull(scope, "scope");
            return scope.GetExtension(ContextId);
        }

        public ScopeExtension EnsureScopeExtension(Scope scope) {
            ContractUtils.RequiresNotNull(scope, "scope");
            ScopeExtension extension = scope.GetExtension(ContextId);

            if (extension == null) {
                extension = CreateScopeExtension(scope);
                if (extension == null) {
                    throw Error.MustReturnScopeExtension();
                }
                return scope.SetExtension(ContextId, extension);
            }

            return extension;
        }

        /// <summary>
        /// Factory for ModuleContext creation. 
        /// It is guaranteed that this method is called once per each ScriptScope the language code is executed within.
        /// </summary>
        public virtual ScopeExtension CreateScopeExtension(Scope scope) {
            return new ScopeExtension(scope);
        }

        #endregion

        #region Source Code Parsing & Compilation

        /// <summary>
        /// Provides a text reader for source code that is to be read from a given stream.
        /// </summary>
        /// <param name="stream">The stream open for reading. The stream must also allow seeking.</param>
        /// <param name="defaultEncoding">An encoding that should be used if the stream doesn't have Unicode or language specific preamble.</param>
        /// <returns>The reader.</returns>
        /// <exception cref="IOException">An I/O error occurs.</exception>
        public virtual SourceCodeReader GetSourceReader(Stream stream, Encoding defaultEncoding) {
            ContractUtils.RequiresNotNull(stream, "stream");
            ContractUtils.RequiresNotNull(defaultEncoding, "defaultEncoding");
            ContractUtils.Requires(stream.CanRead && stream.CanSeek, "stream", "The stream must support reading and seeking");
            
            var result = new StreamReader(stream, defaultEncoding, true);
            result.Peek();
            return new SourceCodeReader(result, result.CurrentEncoding);
        }

        /// <summary>
        /// Creates the language specific CompilerContext object for code compilation.  The 
        /// language should flow any relevant options from the LanguageContext to the 
        /// newly created CompilerContext.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")] // TODO: fix
        public virtual CompilerOptions GetCompilerOptions() {
            return new CompilerOptions();
        }

        /// <summary>
        /// Creates compiler options initialized by the options associated with the module.
        /// </summary>
        public virtual CompilerOptions GetCompilerOptions(Scope scope) {
            return GetCompilerOptions();
        }

        /// <summary>
        /// Parses the source code within a specified compiler context. 
        /// The source unit to parse is held on by the context.
        /// </summary>
        /// <returns><b>null</b> on failure.</returns>
        /// <remarks>Could also set the code properties and line/file mappings on the source unit.</remarks>
        internal protected abstract ScriptCode CompileSourceCode(SourceUnit sourceUnit, CompilerOptions options, ErrorSink errorSink);

        internal protected virtual ScriptCode LoadCompiledCode(DlrMainCallTarget method, string path) {
            return ScriptCode.Load(method, this, path);
        }

        #endregion


        /// <summary>
        /// Looks up the name in the provided Scope using the current language's semantics.
        /// </summary>
        public virtual bool TryLookupName(CodeContext context, SymbolId name, out object value) {
            if (context.Scope.TryLookupName(this, name, out value) && value != Uninitialized.Instance) {
                return true;
            }

            return TryLookupGlobal(context, name, out value);
        }

        /// <summary>
        /// Looks up the name in the provided scope using the current language's semantics.
        /// 
        /// If the name cannot be found throws the language appropriate exception or returns
        /// the language's appropriate default value.
        /// </summary>
        public virtual object LookupName(CodeContext context, SymbolId name) {
            object value;
            if (!TryLookupName(context, name, out value) || value == Uninitialized.Instance) {
                throw MissingName(name);
            }

            return value;
        }

        /// <summary>
        /// Attempts to set the name in the provided scope using the current language's semantics.
        /// </summary>
        public virtual void SetName(CodeContext context, SymbolId name, object value) {
            context.Scope.SetName(name, value);
        }

        /// <summary>
        /// Attempts to remove the name from the provided scope using the current language's semantics.
        /// </summary>
        public virtual bool RemoveName(CodeContext context, SymbolId name) {
            return context.Scope.RemoveName(this, name);
        }

        /// <summary>
        /// Attemps to lookup a global variable using the language's semantics called from
        /// the provided Scope.  The default implementation will attempt to lookup the variable
        /// at the host level.
        /// </summary>
        public virtual bool TryLookupGlobal(CodeContext context, SymbolId name, out object value) {
            value = null;
            return false;
        }

        /// <summary>
        /// Called when a lookup has failed and an exception should be thrown.  Enables the 
        /// language context to throw the appropriate exception for their language when
        /// name lookup fails.
        /// </summary>
        protected internal virtual Exception MissingName(SymbolId name) {
            return Error.NameNotDefined(SymbolTable.IdToString(name));
        }

        /// <summary>
        /// Returns a ModuleGlobalCache for the given name.  
        /// 
        /// This cache enables fast access to global values when a SymbolId is not defined after searching the Scope's.  Usually
        /// a language implements lookup of the global value via TryLookupGlobal.  When GetModuleCache returns a ModuleGlobalCache
        /// a cached value can be used instead of calling TryLookupGlobal avoiding a possibly more expensive lookup from the 
        /// LanguageContext.  The ModuleGlobalCache can be held onto and have its value updated when the cache is invalidated.
        /// 
        /// By default this returns a cache which indicates no caching should occur and the LanguageContext will be 
        /// consulted when a module value is not available. If a LanguageContext only caches some values it can return 
        /// the value from the base method when the value should not be cached.
        /// </summary>
        protected internal virtual ModuleGlobalCache GetModuleCache(SymbolId name) {
            return _noCache;
        }

        /// <summary>
        /// Calls the function with given arguments
        /// </summary>
        /// <param name="context"></param>
        /// <param name="function">The function to call</param>
        /// <param name="args">The argumetns with which to call the function.</param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1716:IdentifiersShouldNotMatchKeywords", MessageId = "Call")] // TODO: fix
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1716:IdentifiersShouldNotMatchKeywords", MessageId = "Function")] // TODO: fix
        public virtual object Call(CodeContext context, object function, object[] args) {
            return null;
        }

        /// <summary>
        /// Calls the function with instance as the "this" value.
        /// </summary>
        /// <param name="context"></param>
        /// <param name="function">The function to call</param>
        /// <param name="instance">The instance to pass as "this".</param>
        /// <param name="args">The rest of the arguments.</param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1716:IdentifiersShouldNotMatchKeywords", MessageId = "Function")] // TODO: fix
        public virtual object CallWithThis(CodeContext context, object function, object instance, object[] args) {
            return null;
        }

        /// <summary>
        /// Calls the function with arguments, extra arguments in tuple and dictionary of keyword arguments
        /// </summary>
        /// <param name="context"></param>
        /// <param name="func">The function to call</param>
        /// <param name="args">The arguments</param>
        /// <param name="names">Argument names</param>
        /// <param name="argsTuple">tuple of extra arguments</param>
        /// <param name="kwDict">keyword dictionary</param>
        /// <returns>The result of the function call.</returns>
        public virtual object CallWithArgsKeywordsTupleDict(CodeContext context, object func, object[] args, string[] names, object argsTuple, object kwDict) {
            return null;
        }

        /// <summary>
        /// Calls function with arguments and additional arguments in the tuple
        /// </summary>
        /// <param name="context"></param>
        /// <param name="func">The function to call</param>
        /// <param name="args">Argument array</param>
        /// <param name="argsTuple">Tuple with extra arguments</param>
        /// <returns>The result of calling the function "func"</returns>
        public virtual object CallWithArgsTuple(CodeContext context, object func, object[] args, object argsTuple) {
            return null;
        }

        /// <summary>
        /// Calls the function with arguments, some of which are keyword arguments.
        /// </summary>
        /// <param name="context"></param>
        /// <param name="func">Function to call</param>
        /// <param name="args">Argument array</param>
        /// <param name="names">Names for some of the arguments</param>
        /// <returns>The result of calling the function "func"</returns>
        public virtual object CallWithKeywordArgs(CodeContext context, object func, object[] args, string[] names) {
            return null;
        }

        // used only by ReflectedEvent.HandlerList
        public virtual bool EqualReturnBool(CodeContext context, object x, object y) {
            return false;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2001:AvoidCallingProblematicMethods", MessageId = "System.Reflection.Assembly.LoadFile")]
        public virtual Assembly LoadAssemblyFromFile(string file) {
#if SILVERLIGHT
            return null;
#else
            return Assembly.LoadFile(file);
#endif
        }

        #region ScriptEngine API

        public virtual Version LanguageVersion {
            get {
                return new Version(0, 0);
            }
        }

        public virtual void SetSearchPaths(ICollection<string> paths) {
            throw new NotSupportedException();
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
        public virtual ICollection<string> GetSearchPaths() {
            return Options.SearchPaths;
        }

#if !SILVERLIGHT
        // Convert a CodeDom to source code, and output the generated code and the line number mappings (if any)
        public virtual SourceUnit GenerateSourceCode(System.CodeDom.CodeObject codeDom, string path, SourceCodeKind kind) {
            throw new NotImplementedException();
        }
#endif

        public virtual TService GetService<TService>(params object[] args) where TService : class {
            return null;
        }

        //TODO these three properties should become abstract and updated for all implementations
        public virtual Guid LanguageGuid {
            get {
                return Guid.Empty;
            }
        }

        public virtual Guid VendorGuid {
            get {
                return Guid.Empty;
            }
        }

        public virtual void Shutdown() {
        }

        public virtual string FormatException(Exception exception) {
            return exception.ToString();
        }

        public virtual Microsoft.Scripting.LanguageOptions Options {
            get {
                return new Microsoft.Scripting.LanguageOptions();
            }
        }

        #region Source Units

        public SourceUnit CreateSnippet(string code, SourceCodeKind kind) {
            return CreateSnippet(code, null, kind);
        }

        public SourceUnit CreateSnippet(string code, string id, SourceCodeKind kind) {
            ContractUtils.RequiresNotNull(code, "code");

            return CreateSourceUnit(new SourceStringContentProvider(code), id, kind);
        }

        public SourceUnit CreateFileUnit(string path) {
            return CreateFileUnit(path, StringUtils.DefaultEncoding);
        }

        public SourceUnit CreateFileUnit(string path, Encoding encoding) {
            return CreateFileUnit(path, encoding, SourceCodeKind.File);
        }

        public SourceUnit CreateFileUnit(string path, Encoding encoding, SourceCodeKind kind) {
            ContractUtils.RequiresNotNull(path, "path");
            ContractUtils.RequiresNotNull(encoding, "encoding");

            TextContentProvider provider = new LanguageBoundTextContentProvider(this, new FileStreamContentProvider(DomainManager.Platform, path), encoding);
            return CreateSourceUnit(provider, path, kind);
        }

        public SourceUnit CreateFileUnit(string path, string content) {
            ContractUtils.RequiresNotNull(path, "path");
            ContractUtils.RequiresNotNull(content, "content");

            TextContentProvider provider = new SourceStringContentProvider(content);
            return CreateSourceUnit(provider, path, SourceCodeKind.File);
        }

        public SourceUnit CreateSourceUnit(StreamContentProvider contentProvider, string path, Encoding encoding, SourceCodeKind kind) {
            ContractUtils.RequiresNotNull(contentProvider, "contentProvider");
            ContractUtils.RequiresNotNull(encoding, "encoding");
            ContractUtils.Requires(path == null || path.Length > 0, "path", Strings.EmptyStringIsInvalidPath);
            ContractUtils.Requires(EnumBounds.IsValid(kind), "kind");
            ContractUtils.Requires(CanCreateSourceCode);

            return new SourceUnit(this, new LanguageBoundTextContentProvider(this, contentProvider, encoding), path, kind);
        }

        public SourceUnit CreateSourceUnit(TextContentProvider contentProvider, string path, SourceCodeKind kind) {
            ContractUtils.RequiresNotNull(contentProvider, "contentProvider");
            ContractUtils.Requires(path == null || path.Length > 0, "path", Strings.EmptyStringIsInvalidPath);
            ContractUtils.Requires(EnumBounds.IsValid(kind), "kind");
            ContractUtils.Requires(CanCreateSourceCode);

            return new SourceUnit(this, contentProvider, path, kind);
        }

        #endregion

        #endregion

        private static T GetArg<T>(object[] arg, int index, bool optional) {
            if (!optional && index >= arg.Length) {
                throw Error.InvalidParamNumForService();
            }

            if (!(arg[index] is T)) {
                throw Error.InvalidArgumentType(String.Format("arg[{0}]", index), typeof(T));
            }

            return (T)arg[index];
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
        public virtual ErrorSink GetCompilerErrorSink() {
            return ErrorSink.Null;
        }

        public virtual void GetExceptionMessage(Exception exception, out string message, out string errorTypeName) {
            message = exception.Message;
            errorTypeName = exception.GetType().Name;
        }

        public virtual int ExecuteProgram(SourceUnit program) {
            ContractUtils.RequiresNotNull(program, "program");

            object returnValue = program.Execute();

            CodeContext context = new CodeContext(new Scope(), this);

            CallSite<Func<CallSite, CodeContext, object, object>> site =
                CallSite<Func<CallSite, CodeContext, object, object>>.Create(OldConvertToAction.Make(Binder, typeof(int), ConversionResultKind.ExplicitTry));

            object exitCode = site.Target(site, context, returnValue);
            return (exitCode != null) ? (int)exitCode : 0;
        }

        #region Object Operations Support

        internal static MetaObject ErrorMetaObject(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
            return onBindingError ?? MetaObject.Throw(target, args, typeof(NotImplementedException), ArrayUtils.EmptyObjects);
        }

        private class DefaultOperationAction : OperationAction {
            internal DefaultOperationAction(string operation)
                : base(operation) {
            }

            public override MetaObject Fallback(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
                return ErrorMetaObject(target, args, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual OperationAction CreateOperationBinder(string operation) {
            return new DefaultOperationAction(operation);
        }

        private class DefaultConvertAction : ConvertAction {
            internal DefaultConvertAction(Type type, bool @explicit)
                : base(type, @explicit) {
            }

            public override MetaObject Fallback(MetaObject self, MetaObject onBindingError) {
                if (ToType.IsAssignableFrom(self.LimitType)) {
                    return new MetaObject(
                        self.Expression,
                        Restrictions.TypeRestriction(self.Expression, self.LimitType)
                    );
                }

                return onBindingError ??
                    MetaObject.Throw(
                        self,
                        MetaObject.EmptyMetaObjects,
                        typeof(ArgumentTypeException),
                        String.Format("Expected {0}, got {1}", ToType.FullName, self.LimitType.FullName)
                    );
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual ConvertAction CreateConvertBinder(Type toType, bool explicitCast) {
            return new DefaultConvertAction(toType, explicitCast);
        }

        private class DefaultGetMemberAction : GetMemberAction {
            internal DefaultGetMemberAction(string name, bool ignoreCase)
                : base(name, ignoreCase) {
            }

            public override MetaObject Fallback(MetaObject self, MetaObject onBindingError) {
                return ErrorMetaObject(self, MetaObject.EmptyMetaObjects, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual GetMemberAction CreateGetMemberBinder(string name, bool ignoreCase) {
            return new DefaultGetMemberAction(name, ignoreCase);
        }

        private class DefaultSetMemberAction : SetMemberAction {
            internal DefaultSetMemberAction(string name, bool ignoreCase)
                : base(name, ignoreCase) {
            }

            public override MetaObject Fallback(MetaObject self, MetaObject value, MetaObject onBindingError) {
                return ErrorMetaObject(self, new MetaObject[] { value }, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual SetMemberAction CreateSetMemberBinder(string name, bool ignoreCase) {
            return new DefaultSetMemberAction(name, ignoreCase);
        }

        private class DefaultDeleteMemberAction : DeleteMemberAction {
            internal DefaultDeleteMemberAction(string name, bool ignoreCase)
                : base(name, ignoreCase) {
            }

            public override MetaObject Fallback(MetaObject self, MetaObject onBindingError) {
                return ErrorMetaObject(self, MetaObject.EmptyMetaObjects, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual DeleteMemberAction CreateDeleteMemberBinder(string name, bool ignoreCase) {
            return new DefaultDeleteMemberAction(name, ignoreCase);
        }

        private class DefaultCallAction : CallAction {
            internal DefaultCallAction(string name, bool ignoreCase, params Argument[] arguments)
                : base(name, ignoreCase, arguments) {
            }

            public override MetaObject Fallback(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
                return ErrorMetaObject(target, args.AddFirst(target), onBindingError);
            }

            public override MetaObject FallbackInvoke(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
                return ErrorMetaObject(target, args.AddFirst(target), onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual CallAction CreateCallBinder(string name, bool ignoreCase, params Argument[] arguments) {
            return new DefaultCallAction(name, ignoreCase, arguments);
        }

        private class DefaultInvokeAction: InvokeAction {
            internal DefaultInvokeAction(params Argument[] arguments)
                : base(arguments) {
            }

            public override MetaObject Fallback(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
                return ErrorMetaObject(target, args, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual InvokeAction CreateInvokeBinder(params Argument[] arguments) {
            return new DefaultInvokeAction(arguments);
        }

        private class DefaultCreateAction : CreateAction {
            internal DefaultCreateAction(params Argument[] arguments)
                : base(arguments) {
            }

            public override MetaObject Fallback(MetaObject target, MetaObject[] args, MetaObject onBindingError) {
                return ErrorMetaObject(target, args, onBindingError);
            }

            public override object HashCookie {
                get { return this; }
            }
        }

        public virtual CreateAction CreateCreateBinder(params Argument[] arguments) {
            return new DefaultCreateAction(arguments);
        }

        #endregion

        /// <summary>
        /// Called by an interpreter when an exception is about to be thrown by an interpreted <see cref="ThrowStatement"/> or
        /// when a CLR method is called that threw an exception.
        /// </summary>
        /// <param name="state">
        /// The current interpreted frame state. The frame is either throwing the exception (via <see cref="ThrowStatement"/>) or 
        /// is the interpreted frame that is calling a CLR method that threw or propagated the exception. 
        /// </param>
        /// <param name="exception">The exception to be (re)thrown.</param>
        /// <param name="isInterpretedThrow">Whether the exception is thrown by an interpreted code (<see cref="ThrowStatement"/>).</param>
        /// <remarks>
        /// The method can be called multiple times for a single exception if the interpreted code calls some CLR code that
        /// calls an interpreted code that throws an exception. The method is called at each interpeted/non-interpreted frame boundary
        /// and in the frame that raised the exception by <see cref="ThrowStatement"/>.
        /// </remarks>
        internal protected virtual void InterpretExceptionThrow(InterpreterState state, Exception exception, bool isInterpretedThrow) {
            Assert.NotNull(state, exception);
            // nop
        }
    }
}
